"""Bezier, a module for creating Bezier curves.
Version 1.1, from < BezierCurveFunction-v1.ipynb > on 2019-05-02
"""

import numpy as np

#__all__ = ["Bezier"]


def twoPoints9(t, P1, P2):
    """
    Returns a point between P1 and P2, parametised by t.
    INPUTS:
        t     float/int; a parameterisation.
        P1    numpy array; a point.
        P2    numpy array; a point.
    OUTPUTS:
        Q1    numpy array; a point.
    """

    if not isinstance(P1, np.ndarray) or not isinstance(P2, np.ndarray):
        raise TypeError('Points must be an instance of the numpy.ndarray!')
    if not isinstance(t, (int, float)):
        raise TypeError('Parameter t must be an int or float!')

    Q1 = (1 - t) * P1 + t * P2
    return Q1

def points9(t, points):
    """
    Returns a list of points interpolated by the Bezier process
    INPUTS:
        t            float/int; a parameterisation.
        points       list of numpy arrays; points.
    OUTPUTS:
        newpoints    list of numpy arrays; points.
    """
    newpoints = []
    #print("points =", points, "\n")
    for i1 in range(0, len(points) - 1):
        #print("i1 =", i1)
        #print("points[i1] =", points[i1])

        newpoints += [twoPoints9(t, points[i1], points[i1 + 1])]
        #print("newpoints  =", newpoints, "\n")
    return newpoints

def point9(t, points):
    """
    Returns a point interpolated by the Bezier process
    INPUTS:
        t            float/int; a parameterisation.
        points       list of numpy arrays; points.
    OUTPUTS:
        newpoint     numpy array; a point.
    """
    newpoints = points
    #print("newpoints = ", newpoints)
    while len(newpoints) > 1:
        newpoints = points9(t, newpoints)
        #print("newpoints in loop = ", newpoints)

    #print("newpoints = ", newpoints)
    #print("newpoints[0] = ", newpoints[0])
    return newpoints[0]

def curve9(t_values, points):
    """
    Returns a point interpolated by the Bezier process
    INPUTS:
        t_values     list of floats/ints; a parameterisation.
        points       list of numpy arrays; points.
    OUTPUTS:
        curve        list of numpy arrays; points.
    """

    if not hasattr(t_values, '__iter__'):
        raise TypeError("`t_values` Must be an iterable of integers or floats, of length greater than 0 .")
    if len(t_values) < 1:
        raise TypeError("`t_values` Must be an iterable of integers or floats, of length greater than 0 .")
    if not isinstance(t_values[0], (int, float)):
        raise TypeError("`t_values` Must be an iterable of integers or floats, of length greater than 0 .")

    curve = np.array([[0.0] * len(points[0])])
    for t in t_values:
        #print("curve                  \n", curve)
        #print("Bezier.Point(t, points) \n", Bezier.Point(t, points))

        curve = np.append(curve, [point9(t, points)], axis=0)

        #print("curve after            \n", curve, "\n--- --- --- --- --- --- ")
    curve = np.delete(curve, 0, 0)
    #print("curve final            \n", curve, "\n--- --- --- --- --- --- ")
    return curve
